"""
Metadata about the package to easily retrieve information about it.
See: https://packaging.python.org/guides/single-sourcing-package-version/
"""

__all__ = [
    "__author__",
    "__copyright__",
    "__license__",
    "__maintainer__",
    "__summary__",
    "__title__",
    "__uri__",
]

__author__ = "Etienne Trimaille"
__maintainer__ = "3Liz"
__copyright__ = f"2021 {__author__}, 2025 3liz"
__license__ = "GNU General Public License v3.0"
__summary__ = "A tool for downloading/managing QGIS plugins from CLI."
__title__ = "QGIS Plugin Manager"
__uri__ = "https://github.com/3liz/qgis-plugin-manager"
